from __future__ import annotations

import typing

T = typing.TypeVar("T")
V = typing.TypeVar("V")

_DATA_PROXY_SLOTS = ("__data_obj__", "__bool_hook__")


class ChainProxy(typing.Generic[T]):
    """
    `ChainProxy` container:
        * stores `data_object`
        * proxying `data_object` attribute value into new `ChainProxy` instance
        when attribute is invoked. If attribute does not exist or attribute value is `None`
        `ChainProxy` instance `data_object` will be `None` and `bool_hook` will be `False`.

    `ChainProxy` instance always returning when attribute is invoked

    :param data_object: Data proxied by `ChainProxy`
    :type data_object: T
    :param: bool_hook: Used in `ChainProxy` instance bool cast.
    bool(chain_proxy_instance) == bool_hook
    :type: bool
    """

    __slots__ = _DATA_PROXY_SLOTS

    def __init__(self, data_object: T, *, bool_hook: bool = False) -> None:
        self.__data_obj__ = data_object
        self.__bool_hook__ = bool_hook

    def __getattr__(self, attr: str) -> ChainProxy[T]:
        object_attribute = getattr(self.__data_obj__, attr, None)

        if object_attribute is None:
            return ChainProxy(
                typing.cast(T, object_attribute),
                bool_hook=False,
            )

        return ChainProxy(object_attribute, bool_hook=True)

    def __get_value__(
        self, *, default: typing.Optional[V] = None
    ) -> typing.Union[T, V]:
        if not self and default is not None:
            return default
        return self.__data_obj__

    def __bool__(self) -> bool:
        return self.__bool_hook__

    def __repr__(self) -> str:
        return f"DataProxy(data_object={self.__data_obj__}, bool_hook={self.__bool_hook__})"

    def __str__(self) -> str:
        return f"data_object={self.__data_obj__}, bool_hook={self.__bool_hook__}"


def _get_value(
    chain_proxy: ChainProxy[T], *, default: typing.Optional[V] = None
) -> typing.Union[T, V]:
    """Protected method for withdrawal object from `ChainProxy`
    :param chain_proxy: `ChainProxy` instance
    :type chain_proxy: ChainProxy[T]
    :param default: Value returned if object is None (default is None)
    :type default: Optional[V]
    :returns: data_object from `ChainProxy`
    :rtype: T or V
    """
    return chain_proxy.__get_value__(default=default)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['safebag']

package_data = \
{'': ['*']}

install_requires = \
['pytest-cov>=3.0.0,<4.0.0']

setup_kwargs = {
    'name': 'safebag',
    'version': '0.1.7',
    'description': 'Little python package implementing optional chaining',
    'long_description': '# Safebag\n[![Stable Version](https://img.shields.io/pypi/v/safebag?color=black)](https://pypi.org/project/safebag/)\n\n_Safebag_ is a little python package implementing optional chaining.\n\n## Installation\n\n```bash\npip install safebag\n```\n\n## Usage\n\n### chain _[[source](https://github.com/galeNightIn/safebag/blob/69e241022b85b3f4566556f3e3e956d5a750eb20/safebag/_methods.py#L9)]_\n\nOptional chain constructor\n\nOptional chain constructed from any object.\n\nChain is used for building sequence of null-safe attribute calls.\n\n```python\nfrom __future__ import annotations\n\nimport dataclasses as dt\nimport typing\n\n\n@dt.dataclass\nclass Node:\n    data: int\n    node: typing.Optional[Node]\n\n\nnodes = Node(data=1, node=Node(data=2, node=None))\n\nfrom safebag import chain\n\nthird_node_proxy = chain(nodes).node.node.node\nprint(third_node_proxy)  # ChainProxy(data_object=None, bool_hook=False)\n```\n\n### get_value _[[source](https://github.com/galeNightIn/safebag/blob/69e241022b85b3f4566556f3e3e956d5a750eb20/safebag/_methods.py#L39)]_\n\nFinal value getter for optional chain.\n\nOptional chain constructed from any object. Chain is used for building sequence of null-safe attribute calls.\n\n```python\nfrom __future__ import annotations\n\nimport dataclasses as dt\nimport typing\n\n\n@dt.dataclass\nclass Node:\n    data: int\n    node: typing.Optional[Node]\n\n\nnodes = Node(data=1, node=Node(data=2, node=None))\n\nfrom safebag import chain, get_value\n\nthird_node_proxy = chain(nodes).node.node.node\nvalue = get_value(third_node_proxy)\nassert value is None\n\nnext_node = chain(nodes).node\nvalue = get_value(next_node)  # Node(data=2, node=None)\n```\n\nUseful in combination with walrus operator:\n\n```python\nif next_node := chain(nodes).node.node:\n    print(get_value(next_node))\n\nif next_node := chain(nodes).node:\n    print(get_value(next_node))  # Node(data=2, node=None)\n```\n\n### ChainProxy _[[source](https://github.com/galeNightIn/safebag/blob/69e241022b85b3f4566556f3e3e956d5a750eb20/safebag/_chain_proxy.py#L11)]_\n\n`ChainProxy` container:\n* stores `data_object`\n* proxying `data_object` attribute value into new `ChainProxy` instance\nwhen attribute is invoked. If attribute does not exist or attribute value is `None`.\n`ChainProxy` instance `data_object` will be `None` and `bool_hook` will be `False`.\n* `ChainProxy` instance always returning when attribute is invoked.',
    'author': 'Alexandr Solovev',
    'author_email': 'nightingale.alex.info@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/galeNightIn/safebag',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import typing

from ._chain_proxy import ChainProxy, _get_value

T = typing.TypeVar("T")
V = typing.TypeVar("V")


def chain(data_object: T) -> ChainProxy[T]:
    """Optional chain constructor

    Optional chain constructed from any object.
    Chain is used for building sequence of null-safe attribute calls.

    :Example:
        >>> from __future__ import annotations
        >>> import dataclasses as dt
        >>> import typing
        >>> @dt.dataclass
        ... class Node:
        ...     data: int
        ...     node: typing.Optional[Node]
        >>> nodes = Node(data=1, node=Node(data=2, node=None))
        >>>
        >>> from safebag import chain
        >>> from safebag import chain, get_value
        >>> third_node_proxy = chain(nodes).node.node.node
        >>> third_node_proxy
        ChainProxy(data_object=None, bool_hook=False)

    :param data_object: Any object for optional chaining
    :type data_object: T
    :return: `ChainProxy` for data_object
    :rtype ChainProxy[T]
    """
    return ChainProxy(data_object)


def get_value(
    proxy_object: ChainProxy, *, default: typing.Optional[V] = None
) -> typing.Union[T, V, None]:
    """Final value getter for optional chain.

    Optional chain constructed from any object.
    Chain is used for building sequence of null-safe attribute calls.

    :Example:
        >>> from __future__ import annotations
        >>> import dataclasses as dt
        >>> import typing
        >>> @dt.dataclass
        ... class Node:
        ...     data: int
        ...     node: typing.Optional[Node]
        >>> nodes = Node(data=1, node=Node(data=2, node=None))
        >>>
        >>> from safebag import chain
        >>> from safebag import chain, get_value
        >>> third_node_proxy = chain(nodes).node.node.node

        >>> value = get_value(third_node_proxy)
        >>> assert value is None

        >>> next_node = chain(nodes).node
        >>> value = get_value(next_node)
        >>> value
        Node(data=2, node=None)

    :Useful in combination with walrus operator:
        >>> if next_node := chain(nodes).node.node:
        ...     print(get_value(next_node))
        >>> if next_node := chain(nodes).node:
        ...     print(get_value(next_node))
        Node(data=2, node=None)
    """
    return _get_value(proxy_object, default=default)
